#include <stdio.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/errno.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>       
#include <signal.h>
#include <getopt.h>

int g_pid;
int g_timeout = 0;
int g_signum = 15;

struct option long_options[] = {
  { "version"          , no_argument,       0, 'V' },
  { "help"             , no_argument,       0, 'h' },
  { "signal"           , required_argument, 0, 's' },
  { 0, 0, 0, 0 } };

void version()
{
  fprintf(stderr, "waitmax version 1.0\n"
	  "Copyright Mathias Kettner 2008\n"
	  "This is free software; see the source for copying conditions.  There is NO\n"
	  "warranty; not even for MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.\n");
  exit(0);
}


void usage()
{
  fprintf(stderr, "Usage: waitmax [-s SIGNUM] MAXTIME PROGRAM [ARGS...]\n"
	  "\n"
	  "Execute PROGRAM as a subprocess. If PROGRAM does not exit before MAXTIME\n"
	  "seconds, it will be killed with SIGTERM or an alternative signal.\n"
	  "\n"
	  "   -s, --signal SIGNUM   kill with SIGNUM on timeout\n"
	  "   -h, --help            this help\n"
	  "   -V, --version         show version an exit\n\n");
  exit(1);
}


void signalhandler(int signum)
{
  if (0 == kill(g_pid, g_signum))
    g_timeout = 1;
}


int main(int argc, char **argv)
{
  int indexptr=0;
  int ret;
  setenv("POSIXLY_CORRECT", "true", 0);
  while (0 <= (ret = getopt_long(argc, argv, "Vhs:", long_options, &indexptr))) {
    switch (ret)
      {
      case 'V':
	version();

      case 'h':
	usage();

      case 's':
	g_signum = strtoul(optarg, 0, 10);
	if (g_signum < 1 || g_signum > NSIG) {
	  fprintf(stderr, "Signalnumber must be between 1 and %d.\n", NSIG);
	  exit(1);
	}
	break;

      default:
	usage(argv[0]);
	exit(1);
	break;
      }
  }

  if (optind + 1 >= argc) usage();

  int maxtime = atoi(argv[optind]);
  if (maxtime <= 0) usage();
  
  g_pid = fork();
  if (g_pid == 0) {
    signal(SIGALRM, signalhandler);
    execvp(argv[optind + 1], argv + optind + 1);
    fprintf(stderr, "Cannot execute %s: %s\n", argv[optind + 1], strerror(errno));
    exit(253);
  }
  
  signal(SIGALRM, signalhandler);
  alarm(maxtime);
  int status;
  while (1) {
      int pid = waitpid(g_pid, &status, 0);
      if (pid <= 0) {
        if (errno == EINTR) continue; // interuppted by alarm
      else
        fprintf(stderr, "Strange: waitpid() fails: %s\n", strerror(errno));
        exit(1);
      }
      else break;
  }
   
  if (WIFEXITED(status)) {
    int exitcode = WEXITSTATUS(status);
    return exitcode;
  }
  else if (WIFSIGNALED(status)) {
    int signum = WTERMSIG(status);
    if (g_timeout)
      return 255;
    else
      return 128 + signum;
  }
  else {
    fprintf(stderr, "Strange: program did neither exit nor was signalled.\n", status);
    return 254;
  }
}
